! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
!+
! Module: binimage_mod
!
! Authors:  Steven L. Snowden - Goddard Space Flight Center (NASA)
!
! Description: Extended Source Analysis Software
!
! Comments:
!
! $Id: binimage_mod.f90,v 1.2 2009/02/20 13:27:08 aibarra Exp $
!-
!  SAS Version 12.0
!  Version 1.1  19 January 2012
!
!  binimage bins images in the XMM-ESAS suite of software.  It operates
!  at the individual image level.
!
!  Input
!      temsc: Threshold for masking
!      idet: Detector selection
!          0: Use the output of comb
!          1: Use a specific exposure
!      prefix: Exposure ID if the detector selection is 1
!      pref: Same as prefix
!      ibin: Binning control
!          1: No binning
!          2: Bin by 2
!          3: Bin by 3
!          4: Bin by 4
!          6: Bin by 6
!          9: Bin by 9
!      iel: Lower energy limit for the energy band in eV
!      ieh: Upper energy limit for the energy band in eV
!      ipart: Particle background control
!          1: include
!          0: do not include
!      iprot: Soft proton control
!          1: include
!          0: do not include
!      iswcx: Solar wind charge exchange control
!          1: include
!          0: do not include
!      imask: Masking control
!          1: include
!          0: do not include
!      maskfile: Mask file name (if absent defaults to using the 
!          exposure map for masking)
! 
!   Examples
!      binimage thresholdmasking=0.02 detector=0 binning=2 elow=350 
!              ehigh=1250 withpartcontrol=1 withsoftcontrol=1
!      binimage thresholdmasking=0.02 detector=1 prefix=1S001 binning=9 
!              elow=350 ehigh=1250 withpartcontrol=1 withsoftcontrol=1
!      binimage thresholdmasking=0.02 detector=0 binning=4 elow=350 
!              ehigh=1250 withpartcontrol=1 withsoftcontrol=0 
!              maskfile=mask.fits
!
!  OUPUT:
!      binned image

module binimage_mod

  use dal
  use param 
  use errorHandling
  use gofutils_mod  ! from goflib, formerly in marq_mod
  use gofmatrix_mod ! from goflib, formerly in marq_mod
  use gofio_mod     ! from goflib, formerly in subsn (SS)

  implicit none 
  private

  public binimage

contains

  !---------------------------------------------------------------------
  subroutine binimage
  !---------------------------------------------------------------------

        integer :: blockn, detect, i, i2, ibin, idet, iel, ieh, iex, &
            iflag, ii, ini4(900,900), isize, j, j1, j2, jj, jj1, jj2, &
            lun
        integer :: status = 0
        real ::  b, back(900,900), bscale, bzero, c, counts(900,900), &
            countso(900,900), e, ex, expose(900,900), mask(900,900), & 
            ootsca, prot(900,900), r, rex, rnb, rnc, rnp, rns, & 
            s, sca, swcx(900,900), temsc, tlim, xcen, ycen
        real, dimension(:,:), allocatable :: output, sigma
        double precision cdelt(2), crpix(2), crval(2)

        logical :: clobber, ipart, iprot, iswcx, warndie

        character (len=1) :: pre(4), pref(5)
        character (len=4) :: prefi
        character (len=5) :: prefix
        character (len=8) :: ctype(2), units
        character (len=25) :: mode
        character (len=60) :: comment
        character (len=80) :: com
        character (len=132) :: infile, infileo, maskfile, & 
            ratefile, sigfile

        equivalence (prefi,pre)
        equivalence (prefix,pref)

        temsc = realParameter("thresholdmasking")
        idet = intParameter("detector")
        ibin = intParameter("binning")
        iel = intParameter("elow")
        ieh = intParameter("ehigh")
        ipart = booleanParameter("withpartcontrol")
        iprot = booleanParameter("withsoftcontrol")
        iswcx = booleanParameter("withswcxcontrol")
        prefix = stringParameter("prefix")
        maskfile = stringParameter("maskfile")
        clobber = booleanParameter("clobber")

        write(errstr,*) "Fractional limit for template: ", temsc
        call message(AppMsg,Always)

!  Process the detector control

        write(errstr,*) "Detector control: ", idet
        call message(AppMsg,Always)
        IF(IDET .GT. 0) THEN
            PRINT *, 'Prefix: ', PREFIX
            IF((PREF(1) .EQ. '1') .OR. (PREF(1) .EQ. '2')) THEN
                DETECT = 1
            ELSE
                DETECT = 2
                DO I=1,4
                    PRE(I) = PREF(I)
                ENDDO
            ENDIF
        ELSE
            DETECT = 0
        ENDIF
        write(errstr,*) "Detector selection: ", detect
        call message(AppMsg,Always)
        write(errstr,*) "Binning factor: ", ibin
        call message(AppMsg,Always)
        write(errstr,*) "Band energies in eV: ", iel, ieh
        call message(AppMsg,Always)
        write(errstr,*) "Particle Background Control: ", ipart
        call message(AppMsg,Always)
        write(errstr,*) "Soft Proton background Control: ", iprot
        call message(AppMsg,Always)
        write(errstr,*) "Solar Wind Charge Exchange background: ", &
                "Control: ", iswcx
        call message(AppMsg,Always)

!  Enter the count array file name

 1000   FORMAT(A110)

        IF(IDET .GT. 0) THEN
            IF(DETECT .EQ. 1) THEN
                IF(IEL .GE. 1000) THEN
                    IF(IEH .GE. 10000) THEN
                        WRITE(INFILE,3001) PREFIX, IEL, IEH
 3001                   FORMAT('mos',A5,'-obj-im-',I4,'-',I5,'.fits')
                    ELSE
                        WRITE(INFILE,3002) PREFIX, IEL, IEH
 3002                   FORMAT('mos',A5,'-obj-im-',I4,'-',I4,'.fits')
                    ENDIF
                ELSEIF(IEH .LT. 1000) THEN
                    WRITE(INFILE,3003) PREFIX, IEL, IEH
 3003               FORMAT('mos',A5,'-obj-im-',I3,'-',I3,'.fits')
                ELSEIF(IEH .GE. 10000) THEN
                    WRITE(INFILE,3004) PREFIX, IEL, IEH
 3004               FORMAT('mos',A5,'-obj-im-',I3,'-',I5,'.fits')
                ELSE
                    WRITE(INFILE,3005) PREFIX, IEL, IEH
 3005               FORMAT('mos',A5,'-obj-im-',I3,'-',I4,'.fits')
                ENDIF
            ELSE
                IF(IEL .GE. 1000) THEN
                    IF(IEH .GE. 10000) THEN
                        WRITE(INFILE,3006) PREFI, IEL, IEH
 3006                   FORMAT('pn',A4,'-obj-im-',I4,'-',I5,'.fits')
                        WRITE(INFILEO,2906) PREFI, IEL, IEH
 2906                   FORMAT('pn',A4,'-obj-im-',I4,'-',I5,'-oot.fits')
                    ELSE
                        WRITE(INFILE,3007) PREFI, IEL, IEH
 3007                   FORMAT('pn',A4,'-obj-im-',I4,'-',I4,'.fits')
                        WRITE(INFILEO,2907) PREFI, IEL, IEH
 2907                   FORMAT('pn',A4,'-obj-im-',I4,'-',I4,'-oot.fits')
                    ENDIF
                ELSEIF(IEH .LT. 1000) THEN
                    WRITE(INFILE,3008) PREFI, IEL, IEH
 3008               FORMAT('pn',A4,'-obj-im-',I3,'-',I3,'.fits')
                    WRITE(INFILEO,2908) PREFI, IEL, IEH
 2908               FORMAT('pn',A4,'-obj-im-',I3,'-',I3,'-oot.fits')
                ELSEIF(IEH .GE. 10000) THEN
                    WRITE(INFILE,3009) PREFI, IEL, IEH
 3009               FORMAT('pn',A4,'-obj-im-',I3,'-',I5,'.fits')
                    WRITE(INFILEO,2909) PREFI, IEL, IEH
 2909               FORMAT('pn',A4,'-obj-im-',I3,'-',I5,'-oot.fits')
                ELSE
                    WRITE(INFILE,3010) PREFI, IEL, IEH
 3010               FORMAT('pn',A4,'-obj-im-',I3,'-',I4,'.fits')
                    WRITE(INFILEO,2910) PREFI, IEL, IEH
 2910               FORMAT('pn',A4,'-obj-im-',I3,'-',I4,'-oot.fits')
                ENDIF
            ENDIF
        ELSE
            IF(IEL .GE. 1000) THEN
                IF(IEH .GE. 10000) THEN
                    WRITE(INFILE,3011) IEL, IEH
 3011               FORMAT('comb-obj-im-',I4,'-',I4,'.fits')
                ELSE
                    WRITE(INFILE,3012) IEL, IEH
 3012               FORMAT('comb-obj-im-',I4,'-',I4,'.fits')
                ENDIF
            ELSEIF(IEH .LT. 1000) THEN
                WRITE(INFILE,3013) IEL, IEH
 3013           FORMAT('comb-obj-im-',I3,'-',I3,'.fits')
            ELSEIF(IEH .GE. 10000) THEN
                WRITE(INFILE,3014) IEL, IEH
 3014           FORMAT('comb-obj-im-',I3,'-',I5,'.fits')
            ELSE
                WRITE(INFILE,3015) IEL, IEH
 3015           FORMAT('comb-obj-im-',I3,'-',I4,'.fits')
            ENDIF
        ENDIF

        CALL RETFIT (INFILE,COUNTS,INI4,1,900, &
                900,CTYPE,CRVAL,CDELT,CRPIX,STATUS)
        write(errstr,*) "BIN_IMAGE: Retrieving count image:", &
                trim(infile)
        call message(AppMsg,Sparse)
        if(status .ne. 0) then
            write(errstr,*) "BIN_IMAGE 01: STATUS=",status, &
                    " on return from retfit"
            call message(AppMsg,Always)
            call error ("noRet","Error reading image")
        endif
        SCA = -1.0/(CDELT(1)*CDELT(2))
        IF(DETECT .EQ. 2) THEN
            CALL RETFIT (INFILEO,COUNTSO,INI4,1,900, &
                    900,CTYPE,CRVAL,CDELT,CRPIX,STATUS)
            write(errstr,*) "BIN_IMAGE: Retrieving OOT count image:", &
                    trim(infile)
            call message(AppMsg,Sparse)
            if(status .ne. 0) then
                write(errstr,*) "BIN_IMAGE 02: STATUS=",status, &
                        " on return from retfit"
                call message(AppMsg,Always)
                call error ("noRet","Error reading image")
            endif

!  This is a bit painful, but we need to know the SUBMODE
!  keyword for the proper scaling of the OOT events

!  Get an unused lun

            call ftgiou(lun,status)
            if(status .ne. 0) then
                write(errstr,*) "BIN_IMAGE 03: STATUS=",status, &
                    " on return from ftgiou"
                call message(AppMsg,Always)
                call error ("noLUN","Cannot aquire LUN")
            endif

!  Open the FITS file

            write(errstr,*) "BIN_IMAGE: Opening image:", trim(infile)
            call message(AppMsg,Sparse)
            call ftopen (lun,infileo,0,blockn,status)
            if(status .ne. 0) then
                write(errstr,*) "BIN_IMAGE 04: STATUS=",status, &
                    " on return from ftopen"
                call message(AppMsg,Always)
                call error ("noOpen","Cannot open image file")
            endif

!  Read in the mode information if a PN observation

            call ftgkys (lun,'SUBMODE',mode,com,status)
            if(status .ne. 0) then
                write(errstr,*) "BIN_IMAGE 05: STATUS=",status, &
                    " on return from ftgkys"
                call message(AppMsg,Always)
                call error ("noKey","Cannot aquire SUBMODE keyword")
             else
                call getootsca(mode,ootsca,warndie)
            endif
            write(errstr,*) "BIN_IMAGE: OOT scale factor:", ootsca
            call message(AppMsg,Sparse)

!  Close the FITS file

            call ftclos(lun,status)
            if(status .ne. 0) then
                write(errstr,*) "BIN_IMAGE 06: STATUS=",status, &
                    " on return from ftclos"
                call message(AppMsg,Always)
                call error ("noClose","Cannot close image file")
            endif
            DO I=1,900
                DO II=1,900
                    COUNTS(I,II) = COUNTS(I,II) - OOTSCA*COUNTSO(I,II)
                ENDDO
            ENDDO
 
!  Release the lun

            call ftfiou(lun,status)
            if(status .ne. 0) then
                write(errstr,*) "BIN_IMAGE 06a: STATUS=",status, &
                    " on return from ftgiou"
                call message(AppMsg,Always)
                call error ("noLUN","Cannot release LUN")
            endif
        endif

!  Enter the exposure array file name

        IF(IDET .GT. 0) THEN
            IF(DETECT .EQ. 1) THEN
                IF(IEL .GE. 1000) THEN
                    IF(IEH .GE. 10000) THEN
                        WRITE(INFILE,3101) PREFIX, IEL, IEH
 3101                   FORMAT('mos',A5,'-exp-im-',I4,'-',I5,'.fits')
                    ELSE
                        WRITE(INFILE,3102) PREFIX, IEL, IEH
 3102                   FORMAT('mos',A5,'-exp-im-',I4,'-',I4,'.fits')
                    ENDIF
                ELSEIF(IEH .LT. 1000) THEN
                    WRITE(INFILE,3103) PREFIX, IEL, IEH
 3103               FORMAT('mos',A5,'-exp-im-',I3,'-',I3,'.fits')
                ELSEIF(IEH .GE. 10000) THEN
                    WRITE(INFILE,3104) PREFIX, IEL, IEH
 3104               FORMAT('mos',A5,'-exp-im-',I3,'-',I5,'.fits')
                ELSE
                    WRITE(INFILE,3105) PREFIX, IEL, IEH
 3105               FORMAT('mos',A5,'-exp-im-',I3,'-',I4,'.fits')
                ENDIF
            ELSE
                IF(IEL .GE. 1000) THEN
                    IF(IEH .GE. 10000) THEN
                        WRITE(INFILE,3106) PREFI, IEL, IEH
 3106                   FORMAT('pn',A4,'-exp-im-',I4,'-',I5,'.fits')
                    ELSE
                        WRITE(INFILE,3107) PREFI, IEL, IEH
 3107                   FORMAT('pn',A4,'-exp-im-',I4,'-',I4,'.fits')
                    ENDIF
                ELSEIF(IEH .LT. 1000) THEN
                    WRITE(INFILE,3108) PREFI, IEL, IEH
 3108               FORMAT('pn',A4,'-exp-im-',I3,'-',I3,'.fits')
                ELSEIF(IEH .GE. 10000) THEN
                    WRITE(INFILE,3109) PREFI, IEL, IEH
 3109               FORMAT('pn',A4,'-exp-im-',I3,'-',I5,'.fits')
                ELSE
                    WRITE(INFILE,3110) PREFI, IEL, IEH
 3110               FORMAT('pn',A4,'-exp-im-',I3,'-',I4,'.fits')
                ENDIF
            ENDIF

        ELSE
            IF(IEL .GE. 1000) THEN
                IF(IEH .GE. 10000) THEN
                    WRITE(INFILE,3111) IEL, IEH
 3111               FORMAT('comb-exp-im-',I4,'-',I5,'.fits')
                ELSE
                    WRITE(INFILE,3112) IEL, IEH
 3112               FORMAT('comb-exp-im-',I4,'-',I4,'.fits')
                ENDIF
            ELSEIF(IEH .LT. 1000) THEN
                WRITE(INFILE,3113) IEL, IEH
 3113           FORMAT('comb-exp-im-',I3,'-',I3,'.fits')
            ELSEIF(IEH .GE. 10000) THEN
                WRITE(INFILE,3114) IEL, IEH
 3114           FORMAT('comb-exp-im-',I3,'-',I5,'.fits')
            ELSE
                WRITE(INFILE,3115) IEL, IEH
 3115           FORMAT('comb-exp-im-',I3,'-',I4,'.fits')
            ENDIF
        ENDIF

        CALL RETFIT (INFILE,EXPOSE,INI4,1,900, &
                900,CTYPE,CRVAL,CDELT,CRPIX,STATUS)
        write(errstr,*) "BIN_IMAGE: Retrieving exposure image:", &
                trim(infile)
        call message(AppMsg,Sparse)
        if(status .ne. 0) then
            write(errstr,*) "BIN_IMAGE 07: STATUS=",status, &
                    " on return from retfit"
            call message(AppMsg,Always)
            call error ("noRet","Error reading exposure image")
        endif

!  Enter the background count array file name

        IF(IPART) THEN
            IF(IDET .GT. 0) THEN
                IF(DETECT .EQ. 1) THEN
                    IF(IEL .GE. 1000) THEN
                        IF(IEH .GE. 10000) THEN
                            WRITE(INFILE,3201) PREFIX, IEL, IEH
 3201                       FORMAT('mos',A5,'-back-im-sky-',I4,'-', &
                                    I5,'.fits')
                        ELSE
                            WRITE(INFILE,3202) PREFIX, IEL, IEH
 3202                       FORMAT('mos',A5,'-back-im-sky-',I4,'-', &
                                    I4,'.fits')
                        ENDIF
                    ELSEIF(IEH .LT. 1000) THEN
                        WRITE(INFILE,3203) PREFIX, IEL, IEH
 3203                   FORMAT('mos',A5,'-back-im-sky-',I3,'-', &
                                I3,'.fits')
                    ELSEIF(IEH .GE. 10000) THEN
                        WRITE(INFILE,3204) PREFIX, IEL, IEH
 3204                   FORMAT('mos',A5,'-back-im-sky-',I3,'-', &
                                I5,'.fits')
                    ELSE
                        WRITE(INFILE,3205) PREFIX, IEL, IEH
 3205                   FORMAT('mos',A5,'-back-im-sky-',I3,'-', &
                                I4,'.fits')
                    ENDIF
                ELSE
                    IF(IEL .GE. 1000) THEN
                        IF(IEH .GE. 10000) THEN
                            WRITE(INFILE,3206) PREFI, IEL, IEH
 3206                       FORMAT('pn',A4,'-back-im-sky-',I4,'-', &
                                    I5,'.fits')
                        ELSE
                            WRITE(INFILE,3207) PREFI, IEL, IEH
 3207                       FORMAT('pn',A4,'-back-im-sky-',I4,'-', &
                                    I4,'.fits')
                        ENDIF
                    ELSEIF(IEH .LT. 1000) THEN
                        WRITE(INFILE,3208) PREFI, IEL, IEH
 3208                   FORMAT('pn',A4,'-back-im-sky-',I3,'-', &
                                I3,'.fits')
                    ELSEIF(IEH .GE. 10000) THEN
                        WRITE(INFILE,3209) PREFI, IEL, IEH
 3209                   FORMAT('pn',A4,'-back-im-sky-',I3,'-', &
                                I5,'.fits')
                    ELSE
                        WRITE(INFILE,3210) PREFI, IEL, IEH
 3210                   FORMAT('pn',A4,'-back-im-sky-',I3,'-', &
                                I4,'.fits')
                    ENDIF
                ENDIF
            ELSE
                IF(IEL .GE. 1000) THEN
                    IF(IEH .GE. 10000) THEN
                        WRITE(INFILE,3211) IEL, IEH
 3211                   FORMAT('comb-back-im-sky-',I4,'-',I5,'.fits')
                    ELSE
                        WRITE(INFILE,3212) IEL, IEH
 3212                   FORMAT('comb-back-im-sky-',I4,'-',I4,'.fits')
                    ENDIF
                ELSEIF(IEH .LT. 1000) THEN
                    WRITE(INFILE,3213) IEL, IEH
 3213               FORMAT('comb-back-im-sky-',I3,'-',I3,'.fits')
                ELSEIF(IEH .GE. 10000) THEN
                    WRITE(INFILE,3214) IEL, IEH
 3214               FORMAT('comb-back-im-sky-',I3,'-',I5,'.fits')
                ELSE
                    WRITE(INFILE,3215) IEL, IEH
 3215               FORMAT('comb-back-im-sky-',I3,'-',I4,'.fits')
                ENDIF
            ENDIF
        ELSE
            INFILE = '    '
        ENDIF
        IF(INFILE .NE. '    ') THEN
            CALL RETFIT (INFILE,BACK,INI4,1,900, &
                        900,CTYPE,CRVAL,CDELT,CRPIX,STATUS)
            write(errstr,*) "BIN_IMAGE: Retrieving QPB image:", &
                    trim(infile)
            call message(AppMsg,Sparse)
            if(status .ne. 0) then
                write(errstr,*) "BIN_IMAGE 08: STATUS=",status, &
                        " on return from retfit"
                call message(AppMsg,Always)
                call error ("noRet","Error reading QPB image")
            endif
        ELSE
            DO I=1,900
                DO II=1,900
                    BACK(I,II) = 0.0
                ENDDO
            ENDDO
        ENDIF

!  Enter the proton count array file name

        IF(IPROT) THEN
            IF(IDET .GT. 0) THEN
                IF(DETECT .EQ. 1) THEN
                    IF(IEL .GE. 1000) THEN
                        IF(IEH .GE. 10000) THEN
                            WRITE(INFILE,3301) PREFIX, IEL, IEH
 3301                       FORMAT('mos',A5,'-prot-im-sky-',I4,'-', &
                                    I5,'.fits')
                        ELSE
                            WRITE(INFILE,3302) PREFIX, IEL, IEH
 3302                       FORMAT('mos',A5,'-prot-im-sky-',I4,'-', &
                                    I4,'.fits')
                        ENDIF
                    ELSEIF(IEH .LT. 1000) THEN
                        WRITE(INFILE,3303) PREFIX, IEL, IEH
 3303                   FORMAT('mos',A5,'-prot-im-sky-',I3,'-', &
                                I3,'.fits')
                    ELSEIF(IEH .GE. 10000) THEN
                        WRITE(INFILE,3304) PREFIX, IEL, IEH
 3304                   FORMAT('mos',A5,'-prot-im-sky-',I3,'-', &
                                I5,'.fits')
                    ELSE
                        WRITE(INFILE,3305) PREFIX, IEL, IEH
 3305                   FORMAT('mos',A5,'-prot-im-sky-',I3,'-', &
                                I4,'.fits')
                    ENDIF
                ELSE
                    IF(IEL .GE. 1000) THEN
                        IF(IEH .GE. 10000) THEN
                            WRITE(INFILE,3306) PREFI, IEL, IEH
 3306                       FORMAT('pn',A4,'-prot-im-sky-',I4,'-', &
                                    I5,'.fits')
                        ELSE
                            WRITE(INFILE,3307) PREFI, IEL, IEH
 3307                       FORMAT('pn',A4,'-prot-im-sky-',I4,'-', &
                                    I4,'.fits')
                        ENDIF
                    ELSEIF(IEH .LT. 1000) THEN
                        WRITE(INFILE,3308) PREFI, IEL, IEH
 3308                   FORMAT('pn',A4,'-prot-im-sky-',I3,'-', &
                                I3,'.fits')
                    ELSEIF(IEH .GE. 10000) THEN
                        WRITE(INFILE,3309) PREFI, IEL, IEH
 3309                   FORMAT('pn',A4,'-prot-im-sky-',I3,'-', &
                                I5,'.fits')
                    ELSE
                        WRITE(INFILE,3310) PREFI, IEL, IEH
 3310                   FORMAT('pn',A4,'-prot-im-sky-',I3,'-', &
                                I4,'.fits')
                    ENDIF
                ENDIF
            ELSE
                IF(IEL .GE. 1000) THEN
                    IF(IEH .GE. 10000) THEN
                        WRITE(INFILE,3311) IEL, IEH
 3311                   FORMAT('comb-prot-im-sky-',I4,'-',I5,'.fits')
                    ELSE
                        WRITE(INFILE,3312) IEL, IEH
 3312                   FORMAT('comb-prot-im-sky-',I4,'-',I4,'.fits')
                    ENDIF
                ELSEIF(IEH .LT. 1000) THEN
                    WRITE(INFILE,3313) IEL, IEH
 3313               FORMAT('comb-prot-im-sky-',I3,'-',I3,'.fits')
                ELSEIF(IEH .GE. 10000) THEN
                    WRITE(INFILE,3314) IEL, IEH
 3314               FORMAT('comb-prot-im-sky-',I3,'-',I5,'.fits')
                ELSE
                    WRITE(INFILE,3315) IEL, IEH
 3315               FORMAT('comb-prot-im-sky-',I3,'-',I4,'.fits')
                ENDIF
            ENDIF
        ELSE
            INFILE = '    '
        ENDIF
        IF(INFILE .NE. '    ') THEN
            CALL RETFIT (INFILE,PROT,INI4,1,900, &
                        900,CTYPE,CRVAL,CDELT,CRPIX,STATUS)
            write(errstr,*) "BIN_IMAGE: Retrieving SP image:", &
                    trim(infile)
            call message(AppMsg,Sparse)
            if(status .ne. 0) then
                write(errstr,*) "BIN_IMAGE 09: STATUS=",status, &
                        " on return from retfit"
                call message(AppMsg,Always)
                call error ("noRet","Error reading SP image")
            endif
        ELSE
            DO I=1,900
                DO II=1,900
                    PROT(I,II) = 0.0
                ENDDO
            ENDDO
        ENDIF

!  Enter the solar wind charge exchange count array file name

        IF(ISWCX) THEN
            IF(IDET .GT. 0) THEN
                IF(DETECT .EQ. 1) THEN
                    IF(IEL .GE. 1000) THEN
                        IF(IEH .GE. 10000) THEN
                            WRITE(INFILE,3401) PREFIX, IEL, IEH
 3401                       FORMAT('mos',A5,'-swcx-im-sky-',I4,'-', &
                                    I5,'.fits')
                        ELSE
                            WRITE(INFILE,3402) PREFIX, IEL, IEH
 3402                       FORMAT('mos',A5,'-swcx-im-sky-',I4,'-', &
                                    I4,'.fits')
                        ENDIF
                    ELSEIF(IEH .LT. 1000) THEN
                        WRITE(INFILE,3403) PREFIX, IEL, IEH
 3403                   FORMAT('mos',A5,'-swcx-im-sky-',I3,'-', &
                                I3,'.fits')
                    ELSEIF(IEH .GE. 10000) THEN
                        WRITE(INFILE,3404) PREFIX, IEL, IEH
 3404                   FORMAT('mos',A5,'-swcx-im-sky-',I3,'-', &
                                I5,'.fits')
                    ELSE
                        WRITE(INFILE,3405) PREFIX, IEL, IEH
 3405                   FORMAT('mos',A5,'-swcx-im-sky-',I3,'-', &
                                I4,'.fits')
                    ENDIF
                ELSE
                    IF(IEL .GE. 1000) THEN
                        IF(IEH .GE. 10000) THEN
                            WRITE(INFILE,3406) PREFI, IEL, IEH
 3406                       FORMAT('pn',A4,'-swcx-im-sky-',I4,'-', &
                                    I5,'.fits')
                        ELSE
                            WRITE(INFILE,3407) PREFI, IEL, IEH
 3407                       FORMAT('pn',A4,'-swcx-im-sky-',I4,'-', &
                                    I4,'.fits')
                        ENDIF
                    ELSEIF(IEH .LT. 1000) THEN
                        WRITE(INFILE,3408) PREFI, IEL, IEH
 3408                   FORMAT('pn',A4,'-swcx-im-sky-',I3,'-', &
                                I3,'.fits')
                    ELSEIF(IEH .GE. 10000) THEN
                        WRITE(INFILE,3409) PREFI, IEL, IEH
 3409                   FORMAT('pn',A4,'-swcx-im-sky-',I3,'-', &
                                I5,'.fits')
                    ELSE
                        WRITE(INFILE,3410) PREFI, IEL, IEH
 3410                   FORMAT('pn',A4,'-swcx-im-sky-',I3,'-', &
                                I4,'.fits')
                    ENDIF
                ENDIF
            ELSE
                IF(IEL .GE. 1000) THEN
                    IF(IEH .GE. 10000) THEN
                        WRITE(INFILE,3411) IEL, IEH
 3411                   FORMAT('comb-swcx-im-sky-',I4,'-',I5,'.fits')
                    ELSE
                        WRITE(INFILE,3412) IEL, IEH
 3412                   FORMAT('comb-swcx-im-sky-',I4,'-',I4,'.fits')
                    ENDIF
                ELSEIF(IEH .LT. 1000) THEN
                    WRITE(INFILE,3413) IEL, IEH
 3413               FORMAT('comb-swcx-im-sky-',I3,'-',I3,'.fits')
                ELSEIF(IEH .GE. 10000) THEN
                    WRITE(INFILE,3414) IEL, IEH
 3414               FORMAT('comb-swcx-im-sky-',I3,'-',I5,'.fits')
                ELSE
                    WRITE(INFILE,3415) IEL, IEH
 3415               FORMAT('comb-swcx-im-sky-',I3,'-',I4,'.fits')
                ENDIF
            ENDIF
        ELSE
            INFILE = '    '
        ENDIF
        IF(INFILE .NE. '    ') THEN
            CALL RETFIT (INFILE,SWCX,INI4,1,900, &
                        900,CTYPE,CRVAL,CDELT,CRPIX,STATUS)
            write(errstr,*) "BIN_IMAGE: Retrieving SWCX image:", &
                    trim(infile)
            call message(AppMsg,Sparse)
            if(status .ne. 0) then
                write(errstr,*) "BIN_IMAGE 10: STATUS=",status, &
                        " on return from retfit"
                call message(AppMsg,Always)
                call error ("noRet","Error reading SWCX image")
            endif
        ELSE
            DO I=1,900
                DO II=1,900
                    SWCX(I,II) = 0.0
                ENDDO
            ENDDO
        ENDIF

!  Find out how many counts we have to play with and the exposure center

        RNC = 0.0
        RNB = 0.0
        RNP = 0.0
        RNS = 0.0
        REX = 0.0
        EX = 0.0
        DO I=1,900
            DO II=1,900
                RNC = RNC + COUNTS(I,II)
                RNB = RNB + BACK(I,II)
                RNP = RNP + PROT(I,II)
                RNS = RNS + SWCX(I,II)
                REX = REX + EXPOSE(I,II)
                BACK(I,II) = BACK(I,II) + PROT(I,II) + SWCX(I,II)
                IF(EXPOSE(I,II) .GT. EX) THEN
                    EX = EXPOSE(I,II)
                    XCEN = I
                    YCEN = II
                ENDIF
            ENDDO
        ENDDO
        write(errstr,*) "BIN_IMAGE: Total Counts:", RNC
        call message(AppMsg,Sparse)
        write(errstr,*) "BIN_IMAGE: Total Exposure:", REX
        call message(AppMsg,Sparse)
        write(errstr,*) "BIN_IMAGE: Total QPB Counts:", RNB
        call message(AppMsg,Sparse)
        write(errstr,*) "BIN_IMAGE: Total SP Counts:", RNP
        call message(AppMsg,Sparse)
        write(errstr,*) "BIN_IMAGE: Total SWCX Counts:", RNS
        call message(AppMsg,Sparse)
 
!  Read in the template

        IF(MASKFILE .EQ. '      ') THEN

!  Use the exposure maps as the template

            DO I=1,900
                DO II=1,900
                    MASK(I,II) = EXPOSE(I,II)
                ENDDO
            ENDDO
        ELSE

!  Return the mask

            CALL RETFIT (MASKFILE,MASK,INI4,1,900, &
                        900,CTYPE,CRVAL,CDELT,CRPIX,STATUS)
            write(errstr,*) "BIN_IMAGE: Retrieving MASK image:", &
                    trim(maskfile)
            call message(AppMsg,Sparse)
            if(status .ne. 0) then
                write(errstr,*) "BIN_IMAGE 11: STATUS=",status, &
                        " on return from retfit"
                call message(AppMsg,Always)
                call error ("noRet","Error reading MASK image")
            endif
        ENDIF

!  Set the template limit

        J = 0
        TLIM = 0.
        DO I=1,900
            DO II=1,900
                IF(MASK(I,II) .GT. 0) THEN
                    TLIM = TLIM + MASK(I,II)
                    J = J + 1
                ENDIF
            ENDDO
        ENDDO
        TLIM = TEMSC*TLIM/J

!  Set the binning limits

        I2 = 900/IBIN
        iex = mod(900,ibin)/2
        allocate(output(i2,i2),sigma(i2,i2))

!  Bin the data

        write(errstr,*) "BIN_IMAGE: Start the Binning"
        call message(AppMsg,Sparse)
        DO I=1,I2
            J2 = I*IBIN + iex
            J1 = J2 - IBIN + 1
            DO II=1,I2
                JJ2 = II*IBIN + iex
                JJ1 = JJ2 - IBIN + 1
                B = 0.
                C = 0.
                E = 0.
                DO J=J1,J2
                    DO JJ=JJ1,JJ2
                        IF(MASK(J,JJ) .GT. TLIM) THEN
                            B = B + BACK(J,JJ)
                            C = C + COUNTS(J,JJ)
                            E = E + EXPOSE(J,JJ)
                        ENDIF
                    ENDDO
                ENDDO
                IF(E .GT. 0.0) THEN
                    R = SCA*(C - B)/E
                    IF(C .GT. 0.0) THEN
                        S = SCA*SQRT(C)/E
                    ELSE
                        S = 0.0
                    ENDIF
                ELSE
                    R = 0.0
                    S = 0.0
                ENDIF
                OUTPUT(I,II) = R
                SIGMA(I,II) = S
            ENDDO
        ENDDO

!  Open the new FITS file

        IF(IEL .GE. 1000) THEN
            IF(IEH .GE. 10000) THEN
                WRITE(RATEFILE,3600) IEL, IEH
 3600           FORMAT('rate-',I4,'-',I5,'.fits')
                WRITE(SIGFILE,3601) IEL, IEH
 3601           FORMAT('sigma-',I4,'-',I5,'.fits')
            ELSE
                WRITE(RATEFILE,3605) IEL, IEH
 3605           FORMAT('rate-',I4,'-',I4,'.fits')
                WRITE(SIGFILE,3606) IEL, IEH
 3606           FORMAT('sigma-',I4,'-',I4,'.fits')
            ENDIF
        ELSE
            IF(IEH .GE. 10000) THEN
                WRITE(RATEFILE,3610) IEL, IEH
 3610           FORMAT('rate-',I3,'-',I5,'.fits')
                WRITE(SIGFILE,3611) IEL, IEH
 3611           FORMAT('sigma-',I3,'-',I5,'.fits')
            ELSEIF(IEH .GE. 1000) THEN
                WRITE(RATEFILE,3615) IEL, IEH
 3615           FORMAT('rate-',I3,'-',I4,'.fits')
                WRITE(SIGFILE,3616) IEL, IEH
 3616           FORMAT('sigma-',I3,'-',I4,'.fits')
           ELSE
                WRITE(RATEFILE,3620) IEL, IEH
 3620           FORMAT('rate-',I3,'-',I3,'.fits')
                WRITE(SIGFILE,3621) IEL, IEH
 3621           FORMAT('sigma-',I3,'-',I3,'.fits')
            ENDIF
        ENDIF

!  Check to see if there is an existing file
!  Get an unused lun

        call ftgiou(lun,status)
        if(status .ne. 0) then
            write(errstr,*) "BIN_IMAGE 12: STATUS=",status, &
                    " on return from ftgiou"
            call message(AppMsg,Always)
            call error ("noLUN","Cannot aquire LUN")
        endif

        CALL FTOPEN (LUN,RATEFILE,0,BLOCKN,STATUS)
        IF(STATUS .EQ. 0) THEN
            write(errstr,*) "File exists with output file name: ", &
                    trim(ratefile)
            call message(AppMsg,Sparse)
            IF(CLOBBER) THEN
                write(errstr,*) "Clobber the file and continue"
                call message(AppMsg,Sparse)
                CALL FTDELT (LUN,STATUS)
                if(status .ne. 0) then
                    write(errstr,*) "BIN_IMAGE 13: STATUS=",status, &
                            " on return from ftdelt"
                    call message(AppMsg,Always)
                    call error ("badDelete","Error deleting file")
                endif
            ELSE
                write(errstr,*) "BIN_IMAGE 14: Output file: ", &
                        trim(ratefile)
                call message(AppMsg,Always)
                call error ("noClobber","File exists and no clobber")
            ENDIF
        ENDIF
        STATUS = 0

        COMMENT = 'Binned count-rate image'
        IFLAG = 1
        ISIZE = 900/ibin
        CDELT(1) = ibin*CDELT(1)
        CDELT(2) = ibin*CDELT(2)
        CRPIX(1) = isize/2 + 0.5
        CRPIX(2) = isize/2 + 0.5
        UNITS = 'c/s/deg2'
        BZERO = 0.0
        BSCALE = 1.0

        CALL PUTFIT(RATEFILE,OUTPUT,INI4,IFLAG,UNITS, &
                COMMENT,ISIZE,ISIZE,BZERO,BSCALE,CTYPE,CRVAL, &
                CDELT,CRPIX,CLOBBER,STATUS)
        write(errstr,*) "BIN_IMAGE: Writing image:", trim(ratefile)
        call message(AppMsg,Sparse)
        if(status .ne. 0) then
            write(errstr,*) "BIN_IMAGE 15: STATUS=",status, &
                    " on return from putfit"
            call message(AppMsg,Always)
            call error ("noPut","Error writing image")
        endif

!  Check to see if there is an existing file

        COMMENT = 'Binned count-rate uncertainty image'

        CALL FTOPEN (LUN,SIGFILE,0,BLOCKN,STATUS)
        IF(STATUS .EQ. 0) THEN
            write(errstr,*) "File exists with output file name: ", &
                    trim(sigfile)
            call message(AppMsg,Sparse)
            IF(CLOBBER) THEN
                write(errstr,*) "Clobber the file and continue"
                call message(AppMsg,Sparse)
                CALL FTDELT (LUN,STATUS)
                if(status .ne. 0) then
                    write(errstr,*) "BIN_IMAGE 16: STATUS=",status, &
                            " on return from ftdelt"
                    call message(AppMsg,Always)
                    call error ("badDelete","Error deleting file")
                endif
            ELSE
                write(errstr,*) "BIN_IMAGE 17: Output file: ", &
                        trim(ratefile)
                call message(AppMsg,Always)
                call error ("noClobber","File exists and no clobber")
            ENDIF
        ENDIF
        STATUS = 0

        CALL PUTFIT(SIGFILE,SIGMA,INI4,IFLAG,UNITS, &
                COMMENT,ISIZE,ISIZE,BZERO,BSCALE,CTYPE,CRVAL, &
                CDELT,CRPIX,CLOBBER,STATUS)
        write(errstr,*) "BIN_IMAGE: Writing image:", trim(sigfile)
        call message(AppMsg,Sparse)
        if(status .ne. 0) then
            write(errstr,*) "BIN_IMAGE 18: STATUS=",status, &
                    " on return from putfit"
            call message(AppMsg,Always)
            call error ("noPut","Error writing image")
        endif

        return

  end subroutine binimage

end module binimage_mod
